"""
Tests for sanity of auxiliary basis sets
"""

import pytest

from .common_testvars import bs_names, bs_metadata


@pytest.mark.parametrize('basis_name', bs_names)
def test_aux_sanity(basis_name):
    """For all basis sets, check that

       1. All aux basis sets exist
       2. That the role of the aux basis set matches the role in
          the orbital basis
    """

    this_metadata = bs_metadata[basis_name]

    for role, aux in this_metadata['auxiliaries'].items():
        assert aux in bs_metadata
        aux_metadata = bs_metadata[aux]

        assert role == aux_metadata['role']


@pytest.mark.parametrize('basis_name', bs_names)
def test_aux_reverse(basis_name):
    """Make sure all aux basis sets are paired with at least one orbital basis set
    """

    this_metadata = bs_metadata[basis_name]
    r = this_metadata['role']
    if r == 'orbital' or r == 'guess':
        return

    # Find where this basis set is listed as an auxiliary
    found = False
    for k, v in bs_metadata.items():
        aux = v['auxiliaries']
        for ak, av in aux.items():
            if av == basis_name:
                assert ak == r
                found = True

    assert found



import re
from .. import lut
from . import helpers

element_entry_re = re.compile(r'^([a-zA-Z]{1,3}):(.*):(.*)$')
shell_info_re = re.compile(r'^([a-zA-Z])\s+(\d+)\s+(\d+)$')


def read_gbasis(basis_lines):
    '''Reads gbasis-formatted file data and converts it to a dictionary with the
       usual BSE fields

       GBASIS only supports electronic shells (no ecp)

       Note that the gbasis format does not store all the fields we
       have, so some fields are left blank
    '''

    basis_lines = helpers.prune_lines(basis_lines, '!#')

    bs_data = {}

    # The file just contains sections separated headed by
    # a line looking like "Al:aug-cc-pV5+dZ:(21s13p6d4f3g2h) -> [8s7p6d4f3g2h]"
    element_sections = helpers.partition_lines(basis_lines, element_entry_re.match, min_size=4)

    found_basis = set()
    for element_lines in element_sections:
        # First line is element + basis name (already checked)
        # Second is highest angular momentum
        element_sym, basis_name, _ = helpers.parse_line_regex(element_entry_re, element_lines[0],
                                                              "Element entry: sym:basis:pattern")

        element_Z = lut.element_Z_from_sym(element_sym, as_str=True)

        # Add what basis we found
        # We only support one basis per file
        found_basis.add(basis_name)

        element_data = helpers.create_element_data(bs_data, element_Z, 'electron_shells')
        max_am = helpers.parse_line_regex(r'^(\d+)$', element_lines[1], 'Highest AM')

        # Split all the shells based on lines beginning with alpha character
        shell_blocks = helpers.partition_lines(element_lines[2:], lambda x: x[0].isalpha())

        # We know how many blocks there should be
        if (max_am + 1) != len(shell_blocks):
            raise RuntimeError("Different number of blocks for element {}. Expected {}, found {}".format(
                element_sym, max_am + 1, len(shell_blocks)))

        # Now loop over the blocks
        found_am = []

        for shell_lines in shell_blocks:
            shell_am, nprim, ngen = helpers.parse_line_regex(shell_info_re, shell_lines[0], 'Shell: AM, nprim, ngen')
            shell_am = lut.amchar_to_int(shell_am)

            func_type = helpers.function_type_from_am(shell_am, 'gto', 'spherical')

            if len(shell_am) > 1:
                raise RuntimeError("Fused AM not supported by gbasis")

            if shell_am in found_am:
                raise RuntimeError("Duplicate AM for element {}: AM {} already found".format(element_sym, shell_am[0]))

            found_am.append(shell_am)

            if shell_am[0] > max_am:
                raise RuntimeError("Found AM greater than max AM: {}".format(shell_am[0]))

            # Now parse the exponents & coefficients
            exponents, coefficients = helpers.parse_primitive_matrix(shell_lines[1:], nprim=nprim, ngen=ngen)

            shell = {
                'function_type': func_type,
                'region': '',
                'angular_momentum': shell_am,
                'exponents': exponents,
                'coefficients': coefficients
            }

            element_data['electron_shells'].append(shell)

    if len(found_basis) > 1:
        raise RuntimeError("Multiple basis sets in a single file: " + str(found_basis))

    return bs_data



"""
Tests for BSE metadata
"""

import os
import pytest
import glob

from basis_set_exchange import api, curate, fileio
from .common_testvars import data_dir, all_table_files, all_metadata_files


def test_get_metadata():
    '''Test the get_metadata function in the API'''

    api.get_metadata(data_dir)


def test_metadata_uptodate(tmp_path):
    '''Tests that the METADATA.json file is up to date'''

    tmp_path = str(tmp_path)  # Needed for python 3.5
    old_path = os.path.join(data_dir, 'METADATA.json')

    # Create a temporary file
    new_path = os.path.join(tmp_path, 'NEW_METADATA.json')
    curate.create_metadata_file(str(new_path), data_dir)

    old_metadata = fileio.read_metadata(old_path)
    new_metadata = fileio.read_metadata(new_path)
    assert old_metadata == new_metadata


@pytest.mark.parametrize('meta_file_path', all_metadata_files)
def test_basis_metadata_pair1(meta_file_path):
    '''Test that each metadata file is paired with a table basis
       in the same directory
    '''

    bsname = os.path.basename(meta_file_path).split('.')[0]
    meta_subdir = os.path.dirname(meta_file_path)
    test_path = os.path.join(data_dir, meta_subdir, bsname)
    test_files = glob.glob(test_path + '.*.table.json')
    assert len(test_files) > 0


@pytest.mark.parametrize('table_file_path', all_table_files)
def test_basis_metadata_pair2(table_file_path):
    '''Test that each table basis is paired with a metadata file
       in the same directory
    '''

    bsname = os.path.basename(table_file_path).split('.')[0]
    table_subdir = os.path.dirname(table_file_path)
    test_file = os.path.join(table_subdir, bsname) + '.metadata.json'
    assert test_file in all_metadata_files


@pytest.mark.parametrize('bs1,bs2', [('6-311g**', '6-311g(d,p)')])
def test_basis_metadata_duplicate(bs1, bs2):
    bsdata1 = api.get_basis(bs1)
    bsdata2 = api.get_basis(bs2)

    # The names are going to be different
    n1 = bsdata1.pop('name')
    n2 = bsdata2.pop('name')

    assert n1 != n2
    assert bsdata1 == bsdata2



'''
Converts basis set data to a specified output format
'''

import bz2
from .bsejson import write_json
from .nwchem import write_nwchem
from .g94 import write_g94, write_xtron, write_psi4
from .gamess_us import write_gamess_us
from .gamess_uk import write_gamess_uk
from .qchem import write_qchem
from .orca import write_orca
from .turbomole import write_turbomole
from .molpro import write_molpro
from .molcas import write_molcas
from .genbas import write_cfour, write_aces2
from .dalton import write_dalton
from .qcschema import write_qcschema
from .demon2k import write_demon2k
from .pqs import write_pqs
from .cp2k import write_cp2k
from .bsedebug import write_bsedebug
from .bdf import write_bdf

_writer_map = {
    'nwchem': {
        'display': 'NWChem',
        'extension': '.nw',
        'comment': '#',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_nwchem
    },
    'gaussian94': {
        'display': 'Gaussian',
        'extension': '.gbs',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_g94
    },
    'psi4': {
        'display': 'Psi4',
        'extension': '.gbs',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_psi4
    },
    'molcas': {
        'display': 'Molcas',
        'extension': '.molcas',
        'comment': '*',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_molcas
    },
    'qchem': {
        'display': 'Q-Chem',
        'extension': '.qchem',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_qchem
    },
    'orca': {
        'display': 'ORCA',
        'extension': '.orca',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_orca
    },
    'dalton': {
        'display': 'Dalton',
        'extension': '.dalton',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_dalton
    },
    'qcschema': {
        'display': 'QCSchema',
        'extension': '.json',
        'comment': None,
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_qcschema
    },
    'cp2k': {
        'display': 'CP2K',
        'extension': '.cp2k',
        'comment': '#',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_cp2k
    },
    'pqs': {
        'display': 'PQS',
        'extension': '.pqs',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_pqs
    },
    'demon2k': {
        'display': 'deMon2K',
        'extension': '.d2k',
        'comment': '#',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_demon2k
    },
    'gamess_us': {
        'display': 'GAMESS US',
        'extension': '.bas',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_gamess_us
    },
    'turbomole': {
        'display': 'Turbomole',
        'extension': '.tm',
        'comment': '#',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_turbomole
    },
    'gamess_uk': {
        'display': 'GAMESS UK',
        'extension': '.bas',
        'comment': '#',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_gamess_uk
    },
    'molpro': {
        'display': 'Molpro',
        'extension': '.mpro',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_molpro
    },
    'cfour': {
        'display': 'CFOUR',
        'extension': '.c4bas',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_cfour
    },
    'acesii': {
        'display': 'ACES II',
        'extension': '.acesii',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_aces2
    },
    'xtron': {
        'display': 'xTron',
        'extension': '.gbs',
        'comment': '!',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_xtron
    },
    'bsedebug': {
        'display': 'BSE Debug',
        'extension': '.bse',
        'comment': '!',
        'valid': None,
        'function': write_bsedebug
    },
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': None,
        'valid': None,
        'function': write_json
    },
    'bdf': {
        'display': 'BDF',
        'extension': '.bdf',
        'comment': '*',
        'valid': set(['gto', 'gto_cartesian', 'gto_spherical', 'scalar_ecp']),
        'function': write_bdf
    }
}


def write_formatted_basis_str(basis_dict, fmt, header=None):
    '''
    Returns the basis set data as a string representing
    the data in the specified output format
    '''

    # make writers case insensitive
    fmt = fmt.lower()
    if fmt not in _writer_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    writer = _writer_map[fmt]

    # Determine if the writer supports all the types in the basis_dict
    if writer['valid'] is not None:
        ftypes = set(basis_dict['function_types'])
        if not ftypes <= writer['valid']:
            raise RuntimeError('Converter {} does not support all function types: {}'.format(fmt, str(ftypes)))

    # Actually do the conversion
    ret_str = writer['function'](basis_dict)

    # Don't add a header for QCSchema, JSON, etc
    if header is not None and _writer_map[fmt]['comment'] is not None:
        comment_str = _writer_map[fmt]['comment']
        header_str = comment_str + comment_str.join(header.splitlines(True))
        ret_str = header_str + '\n\n' + ret_str

    # HACK - Psi4 requires the first non-comment line be spherical/cartesian
    #        so we have to add that before the header
    if fmt == 'psi4':
        types = basis_dict['function_types']
        harm_type = 'spherical' if 'gto_spherical' in types else 'cartesian'
        ret_str = harm_type + '\n\n' + ret_str

    return ret_str


def write_formatted_basis_file(basis_dict, outfile_path, basis_fmt=None, header=None):
    if basis_fmt is None:
        for k, v in _writer_map.items():
            ext = v['extension']
            ext_bz2 = ext + '.bz2'
            if outfile_path.endswith(ext) or outfile_path.endswith(ext_bz2):
                basis_fmt = k
                break
        else:
            raise RuntimeError("Unable to determine basis set format of '{}'".format(outfile_path))
    else:
        basis_fmt = basis_fmt.lower()

    if basis_fmt not in _writer_map:
        raise RuntimeError("Unknown output file format '{}'".format(basis_fmt))

    basis_str = write_formatted_basis_str(basis_dict, basis_fmt, header)

    if outfile_path.endswith('.bz2'):
        with bz2.open(outfile_path, 'wt') as f:
            f.write(basis_str)
    else:
        with open(outfile_path, 'w') as f:
            f.write(basis_str)


def get_writer_formats(function_types=None):
    '''Return information about the basis set formats available for writing

    The returned data is a map of format to display name. The format
    can be passed as the fmt argument to :func:`write_formatted_basis_str` or
    :func:`write_formatted_basis_file`

    If a list is specified for function_types, only those formats
    supporting the given function types will be returned.
    '''

    if function_types is None:
        return {k: v['display'] for k, v in _writer_map.items()}

    ftypes = [x.lower() for x in function_types]
    ftypes = set(ftypes)
    ret = []

    for fmt, v in _writer_map.items():
        if v['valid'] is None or ftypes <= v['valid']:
            ret.append(fmt)
    return ret


def get_format_extension(fmt):
    '''
    Returns the recommended extension for a given format
    '''

    if fmt is None:
        return 'dict'

    fmt = fmt.lower()
    if fmt not in _writer_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    return _writer_map[fmt]['extension']
